#include "RecognitionPartitions.h"

	//This function generates all the partitions of the set between 1-->N
	//Each subset in the partition is stored as an integer bit mask, with a 
	//vector of these bitmasks forming the partition
void RecognitionPartitions::insertRecognitionPartitions(int N, RecognitionResults recResults)
{
	deque<vector<vector<int> > > allPartitions;
	if(N < 1)
		return;

	partitionMap.clear();
	generatePartitions(N, allPartitions);
	convertPartitionSubsets(allPartitions);
	insertAllPartitionData(allPartitions, recResults);
}


	//For the partition with the highest combined confidence, return in the "out" parameters the names and confidences
void RecognitionPartitions::getHighestPartitionData(vector<int> &highestKey, vector<vector<string> > &names, vector<vector<double> > &confs)
{
	highestKey =  highestCombinedConfKey();

	names = partitionMap[highestKey].names;
	confs = partitionMap[highestKey].confs;
}



//Stores onto the allPartitions data structure all of the set partitions
//from 1->N.  For N = 3, we have:
//{1, 2, 3}; {1, 2}{3}; {1}{2, 3}; {1, 3}{2}; {1}{2}{3}
void RecognitionPartitions::generatePartitions(int N, deque<vector<vector<int> > > &allPartitions)
{
		//Initialize allPartitions with the partitions from 1->1,
		//which is simply the set {1}
	vector<vector<int> > currPart;
	vector<int> currSet;
	currSet.push_back(1);
	currPart.push_back(currSet);
	allPartitions.push_back(currPart);

	for(int i = 2; i <= N; i++)
	{
		currSet[0] = i;
		int initSize = allPartitions.size();  
		
		for(int j = 0; j < initSize; j++)
		{
				//Now apply first part of magic operator -- add i as separate set in each partition
			currPart = allPartitions[j];
			currPart.push_back(currSet);
			allPartitions.push_back(currPart);
							
				//Now apply the second part of the magic operator:
				//for each subset in each partition, add i to the subset
			for(unsigned int k = 0; k < allPartitions[j].size(); k++)
			{
				allPartitions[j][k].push_back(i);
				currPart = allPartitions[j];
				allPartitions.push_back(currPart);
				allPartitions[j][k].erase(allPartitions[j][k].begin() + allPartitions[j][k].size() - 1);
			}
		}

		allPartitions.erase(allPartitions.begin(), allPartitions.begin() + initSize);
	}
}


	//For each subset in each partition, convert the elements into a bit representation
	//where subset = 2 ^ elem1 + 2 ^ elem2 + ...
void RecognitionPartitions::convertPartitionSubsets(deque<vector<vector<int> > > &allPartitions)
{
	//Now that we have all the partitions, combine subsets.
	int amount;
	for(unsigned int i = 0; i < allPartitions.size(); i++)
		for(unsigned int j = 0; j < allPartitions[i].size(); j++)
		{
			amount = 0;
			for(unsigned int k = 0; k < allPartitions[i][j].size(); k++)
			{
				amount += static_cast<int>(pow(2, allPartitions[i][j][k] - 1));
			}
			allPartitions[i][j].clear();
			allPartitions[i][j].push_back(amount);
		}
}


	//For each partition, copy partition elements from the 3-d vector to a 1-d vector
	//and send the 1-d vector to insertPartitionData
void RecognitionPartitions::insertAllPartitionData(const deque<vector<vector<int> > > allPartitions, 
												  RecognitionResults recResults)
{
	vector<int> partition;
	for(unsigned int i = 0; i < allPartitions.size(); i++)
	{
		partition.clear();
		for(unsigned int j = 0; j < allPartitions[i].size(); j++)
		{
			partition.push_back(allPartitions[i][j][0]);
		}

	
		insertPartitionData(partition, recResults);
	}
}


	//Insert into the partition map the data for the subset being passed in
	//using the data in the passed recResults object
void RecognitionPartitions::insertPartitionData(vector<int> subset, RecognitionResults recResults)
{
	PartitionsData data;
	data.combinedConf = 0;

	for(unsigned int i = 0; i < subset.size(); i++)
	{
		data.names.push_back(recResults.getNames(subset[i]));
		data.confs.push_back(recResults.getConfidences(subset[i]));
		data.combinedConf += data.confs[i][0];
	}

		//Using average for the confidence for now.
	data.combinedConf /= subset.size();

	partitionMap[subset] = data;
}


	//Displays all the keys in the map to the screen
	//This function should only be used for TESTING.
void RecognitionPartitions::outputKeys()
{
	for(map<vector<int>, PartitionsData>::iterator iter = partitionMap.begin();
			iter != partitionMap.end(); iter++)
	{
		cout << "Key: ";
		for(unsigned int i = 0; i < iter->first.size(); i++)
			cout << iter->first[i] << " ";
		cout << endl;
	}
	cout << endl << endl;
}


	//Displays all the data for partitions to the screen
	//This function should only be used for TESTING.
void RecognitionPartitions::outputPartitionData()
{
		
	for(map<vector<int>, PartitionsData>::iterator iter = partitionMap.begin();
			iter != partitionMap.end(); iter++)
	{
		cout << "\n\nPartition: ";
		for(unsigned int i = 0; i < iter->first.size(); i++)
			cout << iter->first[i] << " ";

		cout << "\nCombined Confidence: " << iter->second.combinedConf;
		cout << endl;
		for(unsigned int i = 0; i < iter->first.size(); i++)
		{
			cout << "\nSubset: " << iter->first[i];
			cout << "\nNames: ";
			for(unsigned int j = 0; j < iter->second.names[i].size(); j++)
				cout << iter->second.names[i][j] << "   "; 
			
			cout << "\nConfidences: ";
			for(unsigned int j = 0; j < iter->second.confs[i].size(); j++)
				cout << iter->second.confs[i][j];
			
		}

		cout << endl << endl;
	}
}


	//Goes through the partition results and finds the partition with the highest combined conf
	//For this partition, it returns the key to the map that it is represented by.
vector<int> RecognitionPartitions::highestCombinedConfKey()
{
	double highestConf = -1;
	vector<int> highestKey;
	for(map<vector<int>, PartitionsData>::iterator iter = partitionMap.begin(); 
			iter!= partitionMap.end(); iter++)
	{
		if(iter->second.combinedConf > highestConf)
		{
			highestKey = iter->first;
			highestConf = iter->second.combinedConf;
		}
	}


	return highestKey;
}


//----------------------------------------------------------------------------------------------
// Testing Functions
//----------------------------------------------------------------------------------------------

//Returns all the keys to a program who can test them to make sure they are valid.
vector<vector<int> > RecognitionPartitions::returnPartitions(int N)
{
	deque<vector<vector<int> > > allPartitions;
	vector<vector<int> > retParts;
	generatePartitions(N, allPartitions);
	convertPartitionSubsets(allPartitions);

	vector<int> partition;
	for(unsigned int i = 0; i < allPartitions.size(); i++)
	{
		partition.clear();
		for(unsigned int j = 0; j < allPartitions[i].size(); j++)
			partition.push_back(allPartitions[i][j][0]);

		retParts.push_back(partition);
	}

	return retParts;
}


vector<vector<double> > RecognitionPartitions::returnPartConfs(vector<int> partKey)
{
	return partitionMap[partKey].confs;
}

vector<vector<string> > RecognitionPartitions::returnPartNames(vector<int> partKey)
{
	return partitionMap[partKey].names;
}

double RecognitionPartitions::returnCombinedConf(vector<int> partKey)
{
	return partitionMap[partKey].combinedConf;
}
