/**********************************************************************************************************
/RecognitionPartitions.h
/This class stores the set of all partitions of the set from 1->N and the associated combined confidences,
/individual symbol confidences, and recognized symbols (classes) for each partition.
/
/The partition is stored as a vector of integers, where each integer is a bitmask for one subset in the
/partition.  This vector of integers is then mapped on the other data associated with it
/(combined confidence, symbol names, etc).
/Example: Set: 1->5; Partition: {1} {2 3 4} {5}; Bitmask: {1} {2 + 4 + 8 = 14} {16}; associated data
/
/Warning: The algorithm used in generatePartitions is not the fastest algorithm available and when N is 
/about nine or greater, there is a huge performance hit.
/
/This class is fundamentally dependent on the RecognitionResults class (i.e. I did not make it general
/enough to be used in other programs).  I would consider it a fully private class that is a friend of
/recognition results; or, an inner class to RecResults.
**********************************************************************************************************/
#include<vector>
#include<deque>
#include<map>
#include<string>
#include <algorithm>
#include <stdexcept>
#include "RecognitionResults.h"

using std::vector;
using std::string;
using std::map;
using std::deque;

#ifndef __RECOGNITIONPARTITIONS_H__
#define __RECOGNITIONPARTITIONS_H__

class RecognitionPartitions
{
public:
	void insertRecognitionPartitions(int N, RecognitionResults recResults);
	void getHighestPartitionData(vector<int> &highestKey, vector<vector<string> > &names, vector<vector<double> > &confs);

	//These functions are for TESTING purposes ONLY.
	void outputPartitionData();
	void outputKeys();
	vector<vector<int> > returnPartitions(int N);
	vector<vector<double> > returnPartConfs(vector<int> partKey);
	vector<vector<string> > returnPartNames(vector<int> partKey);
	double returnCombinedConf(vector<int> partKey);


private:
	void generatePartitions(int N, deque<vector<vector<int> > > &allPartitions);
	void insertPartitionData(vector<int> subset, RecognitionResults recResults);
	void convertPartitionSubsets(deque<vector<vector<int> > > &allPartitions);
	void insertAllPartitionData(const deque<vector<vector<int> > > allPartitions, RecognitionResults recResults);
	vector<int> highestCombinedConfKey();
		
		//Data structure to store the detailed information for a partition
	struct Partition_T
	{
		vector<vector<double> > confs;
		vector<vector<string> > names;
		double combinedConf;
	};

	typedef struct Partition_T PartitionsData;

		//The key of the map is a vector where each element of the vector
		//represents one subset of the partition and the second part of the map
		//stores a reference to the data that matches each subset.
	map<vector<int>, PartitionsData> partitionMap;
};
#endif
