/* *************************************************************************
* Copyright (c) 1999, Steve Smithies                                       *
*                                                                          *
* This program is free software; you can redistribute it and/or modify it  *
* under the terms of the GNU General Public License as published by the    *
* Free Software Foundation.  See http://www.fsf.org/copyleft/gpl.html      *
*                                                                          *
* This program is distributed in the hope that it will be useful, but      *
* WITHOUT EXPRESS OR IMPLIED WARRANTY of merchantability or fitness for    *
* any particular purpose.  See the GNU General Public License for more     *
* details.                                                                 *
*                                                                          *
* Revision History:                                                        *
*   v 1.0.0 Original Version, Steve Smithies 1999                          *
*   v 1.0.1 R. Zanibbi, March 15, 2002                                     *
*     - CharRec() altered to use new Font_File_Base_Directory() fn.        *
***************************************************************************/

#include <assert.h>
#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <stdio.h>

#include <tcl.h>
#include <tk.h>

#include "charrec.h"

//JZ: May 24, 2006
//Include fstream since we output stroke data to a file
#include <fstream>

CharRec::CharRec() 
{
  
  fprintf(stderr,"Initialising recogniser, please wait...\n");

  // R. Zanibbi March 15, 2002: using the FFES base dir returned
  // by Font_File_Base_Directory.
  //
  // R. Zanibbi April 24, 2003: updating to use revised CIT4.3.1
  //
  //JZ: Aug 18, 2006: Added new recogniser for images
  // RZ Aug 10, 2007: Changed CIT path name.
  char dataDir[256] = "";
  char imageDataDir[256] ="";
  strcat(dataDir,Font_File_Base_Directory());
  strcat(dataDir,"/CIT4.3.2/data");
  strcat(imageDataDir,Font_File_Base_Directory());
  strcat(imageDataDir,"/CIT4.3.2/imagedata");
  MaxStrokeCount = Max_Strokes_In_Font(dataDir);
  Recogniser = new FeatureRecognizer(dataDir);
  ImageRecogniser = new FeatureRecognizer(imageDataDir);
  fprintf(stderr,"Recogniser ready.\n");
}

CharRec::~CharRec()
{
  delete Recogniser;
  delete ImageRecogniser;
}


// return where the font files' tree is based, using the
// environment variable "FONTFILEDIR"
char * CharRec::Font_File_Base_Directory(void)
{
  // R. Zanibbi, March 15, 2002: changed FONTFILEDIR to FFES_DIR
  // so environment path gives FFES directory, not path to data dir.
  char *res = getenv("FFES_DIR");
  if (res == NULL)
  {
	printf("Error: FFES_DIR environment variable is undefined.\nFFES_DIR should be defined as the path to the ffes_0.4 directory.\n");
	exit(1);
  }
  else 
    return(res);
}


// we're now going to print out all the stuff we figured out, and
// let the calling program deal with it as it sees fit
// this kinda parallels the old Update_Rec_Data function

//JZ: Jun 19, 2006
//Update: Now we pass it the recognition results for an entire partition
//and have it return the entire partition's results.
string CharRec::MakeRecognitionResultForTcl(const vector<int> &highestKey, const vector<vector<string> > &names)
{
	stringstream ss;
	//Build string here and display it in cout for now.
	for(unsigned int i = 0; i < highestKey.size(); i++)
	{
		vector<int> acceptedStrokes;

			//First part of output: the strokes used for this character.  1-based (not stroke IDs)
		ss << "{ ";
		int bitMask = 1;
		for(int j = 0; j < StrokeCount; j++, bitMask <<= 1)	//bitMask = 1, 2, 4, 8, ...
			if(highestKey[i] & bitMask)
			{
				acceptedStrokes.push_back(j);
				ss << j + 1 << " ";
			}
		ss << " } { ";
		
			//Second part of output: Stroke IDs
		for(unsigned int j = 0; j < acceptedStrokes.size(); j++)
			for(unsigned int k = 0; k < StrokeIDs[acceptedStrokes[j]].size(); k++)
				ss << " " << StrokeIDs[acceptedStrokes[j]][k];

		ss << " } { ";

			//Third part of output: Symbol names (classes) 
		for(unsigned int j = 0; j < names[i].size(); j++)
			ss << names[i][j] << " ";

		ss << "} ";
	}

	return ss.str();
}



// R. Zanibbi, April 29, 2004
// And a function to return the names of the available classes.
char* CharRec::GetSymbolNames()
{
	return Recogniser -> GetSymbolNames();
}


bool CharRec::MultilinesIntersect(Multiline &M1, Multiline &M2)
{
  if ((M1.Box()*M2.Box()).Area() == 0) // no intersection of bounding boxes, don't waste time doing
    return false;		     // slower overlap testing
 

  
  for (int icompnum=1;icompnum<M1.Points();icompnum++) {
    for (int jcompnum=1;jcompnum<M2.Points();jcompnum++) {
      Position_2D A,B,C,D;
      A.x = M1.Vertex(icompnum-1).X();
      A.y = M1.Vertex(icompnum-1).Y();
      B.x = M1.Vertex(icompnum).X();
      B.y = M1.Vertex(icompnum).Y();

      C.x = M2.Vertex(jcompnum-1).X();
      C.y = M2.Vertex(jcompnum-1).Y();
      D.x = M2.Vertex(jcompnum).X();
      D.y = M2.Vertex(jcompnum).Y();


      if (LineIntersectTest(A,B,C,D)) {
		  return true;
	  }
      
    } // for jcompnum
  } // icompnum


  return false;
}


	// Determine which strokes overlap which other strokes, and combine them
	// Look for intersections between strokes.
void CharRec::CombineOverlappingMultilines(vector<Multiline> &StrokeMultiline)
{
	int ChangeMade;
	int WeOverlap;

	do 
	{
		ChangeMade = false;
		for (int i=0; ((!ChangeMade) && (i<(StrokeCount-1))); i++)
		{
			for (int j=i+1; ((!ChangeMade) && (j<StrokeCount)); j++)
			{
				WeOverlap = MultilinesIntersect(StrokeMultiline[i], StrokeMultiline[j]);

				if (WeOverlap)
				{

					StrokeMultiline[i] += StrokeMultiline[j];
					
					for (unsigned int CollCount = 0; CollCount < StrokeIDs[j].size(); CollCount++) 
						StrokeIDs[i].push_back( StrokeIDs[j].at(CollCount) );

					StrokeMultiline.erase(StrokeMultiline.begin() + j);

					for(int k = j; k < StrokeCount - 1; k++)
					{
						StrokeIDs[k] = StrokeIDs[k + 1];
					}

					StrokeIDs[StrokeCount - 1].clear();
					StrokeCount--;
					ChangeMade = true;
				} else {}
			}
		}
	} while (ChangeMade); // repeat until no more changes are made
}


//JZ: Jun  7, 2006
//inString is a set of integer values separated by spaces
//This function will take each integer value and put it as an element for dest
void tokenizeIntString(const char *inString, vector<int> *dest)
{
		//(JZ) You MUST make a copy of the inString or else data will be corrupted
		//when you return to TCL
	char otherString[strlen(inString)];
	strcpy(otherString,inString);

	char *p;
	p = strtok((otherString), " ");
	while (p)
	{
		dest->push_back(atoi(p));	
		p = strtok(0, " ");
	}
}



//JZ: Jun  7, 2006
//This function takes the data given from TCL
//and converts it to a vector of multilines
//which are stored in the data member StrokeMultilines
vector<Multiline> CharRec::convertTclDataToMultilines(int argc, const char *argv[])
{
		//Convert data from character strings to integers
	int NumberOfMultilines = atoi(argv[1]);
	vector<int> OriginalCollection;
	vector<int> PointCounts;
	vector<int> Points;
	vector<int> Times;

	tokenizeIntString(argv[2], &OriginalCollection);
	tokenizeIntString(argv[3], &PointCounts);
  	tokenizeIntString(argv[4], &Points);
  	tokenizeIntString(argv[5], &Times);

  	StrokeMultiline.clear();
  	StrokeIDs[0].clear();

		//Don't filter any points for now
  	FilterMultilineData(&PointCounts, &Points, &Times, -1);

  	int currentIndex = 0;

	//-------------------------------------------------------------------
	//For each stroke (multiline) store its ID in StrokeIDs
	//and the list of points in StrokeMultiline
	//-------------------------------------------------------------------
  	for (StrokeCount = 0; StrokeCount < NumberOfMultilines; StrokeCount++) 	{
    	StrokeIDs[StrokeCount].clear();      
     	StrokeIDs[StrokeCount].push_back(OriginalCollection[StrokeCount]);

      	Multiline ML;
      	unsigned int PointPairs = PointCounts[StrokeCount];

		//---------------------------------------------------------------
		//Store all the points (and times) of the current multiline on ML 
		//---------------------------------------------------------------
      	for (unsigned int j=0; j < PointPairs;j++)	{
		  	if ((Points[currentIndex] == -1) || (Points[currentIndex] == -1))  {
				ML << Multiline::Jump;
		  	} else {
			    	//Add to multiline (x, y, time)
			  	ML << TimedVec2 (Points[currentIndex],Points[currentIndex + 1] , Times[currentIndex /2]);
			  	currentIndex += 2;
		  	}
	  	} 
		StrokeMultiline.push_back(ML);
  	} 
  	return StrokeMultiline;
}



//JZ: Aug  14, 2006
//This function takes the data given from TCL for an image and converts it to a single multiline
//with a jump in each position where time is = -1
vector<Multiline> CharRec::convertTclImageDataToMultilines(int argc, const char *argv[]) {
		//Convert data from character strings to integers
	int NumberOfMultilines = atoi(argv[1]);
	vector<int> OriginalCollection;
	vector<int> PointCounts;
	vector<int> Points;
	vector<int> Times;

	tokenizeIntString(argv[2], &OriginalCollection);
	tokenizeIntString(argv[3], &PointCounts);
  	tokenizeIntString(argv[4], &Points);
  	tokenizeIntString(argv[5], &Times);

  	StrokeMultiline.clear();
  	StrokeIDs[0].clear();

	int currentIndex = 0;
	int timeAdjust = 0;

		//This function is guaranteed to only have one 
  	for (StrokeCount = 0; StrokeCount < NumberOfMultilines; StrokeCount++) 	{
    	StrokeIDs[StrokeCount].clear();      
     	StrokeIDs[StrokeCount].push_back(OriginalCollection[StrokeCount]);

      	Multiline ML;
      	unsigned int PointPairs = PointCounts[StrokeCount];

		//---------------------------------------------------------------
		//Store all the points (and times) of the current multiline on ML 
		//---------------------------------------------------------------
      	for (unsigned int j=0; j < PointPairs;j++)	{
		  	if ((Points[currentIndex] == -1) || (Points[currentIndex] == -1) || 
					Times[currentIndex / 2 + timeAdjust] == -1)  {
				ML << Multiline::Jump;
				timeAdjust++;
		  	} else {
			    	//Add to multiline (x, y, time)
			  	ML << TimedVec2 (Points[currentIndex],Points[currentIndex + 1], Times[currentIndex / 2 + timeAdjust]);
			  	currentIndex += 2;
		  	}
	  	} 
		StrokeMultiline.push_back(ML);
  	}
  	return StrokeMultiline;
}


//JZ: Jun  8, 2006
//This function is here to allow me to put my function that calls recognize
//in the segmenter.cc file
FeatureRecognizer* CharRec::getRecogniser()
{
	return Recogniser;
}


//JZ: Aug 18, 2006
//To get the new image recogniser
FeatureRecognizer* CharRec::getImageRecogniser()
{
	return ImageRecogniser;
}
