/* *************************************************************************
* Copyright (c) 1999, Steve Smithies                                       *
*                                                                          *
* This program is free software; you can redistribute it and/or modify it  *
* under the terms of the GNU General Public License as published by the    *
* Free Software Foundation.  See http://www.fsf.org/copyleft/gpl.html      *
*                                                                          *
* This program is distributed in the hope that it will be useful, but      *
* WITHOUT EXPRESS OR IMPLIED WARRANTY of merchantability or fitness for    *
* any particular purpose.  See the GNU General Public License for more     *
* details.                                                                 *
*                                                                          *
*   HISTORY:                                                               *
*     Original Version, 1999                                               *
*     March 4, 2002, Richard Zanibbi                                       *
*       -commented out includes	of latex_preview.h and                     *
*        intersect.h; these are unused, and have been removed              *
*        from the distribution; Do_LaTeX_Preview function                  *
*        has also been removed.                                            *
*       -thumb.h and corresponding thumbnail code has                      *
*        also been commented out and removed.                              *
*     March 15, 2002, Richard Zanibbi                                      *
*       - changed my_init to use an environment variable to set the        *
*         base path for FFES again.  Also removed warning about            *
*         FONTFILEDIR in main (now using FFES_DIR: see common.cc)          *
*     April 23, 2004, Richard Zanibbi                                      *
*     	- changed path to FFES Tk interface file.                          *
*     	- added code to tell the user if FFES_DIR is incorrect.            *
*     April 29, 2004, R. Zanibbi                                           *
*       - added functions to report the number and the names of CIT symbols*
*	  May 21, 2004 R. Zanibbi                                              *
*		- changes to last arg of callbacks (now "const"), and resulting    *
*         new variables in the "strtok" calls in "recognise_this."         *
***************************************************************************/

/******************************************************************************
 * RESEARCH PUBLICATIONS ABOUT FFES:
 
	1. Steve Smithies. Freehand Formula Entry System. Master's thesis, 
	   University of Otago, Dunedin, New Zealand, May 1999. 
	   
	2. S. Smithies, K. Novins, J. Arvo. "A handwriting-based equation editor," 
	   In Proc. Graphics Interface, Kingston, Ontario, Canada, June 1999. 
	   
	3. S. Smithies, K. Novins, and J. Arvo. Equation entry and editing via 
	   handwriting and gesture recognition. Behaviour and 
	   Information Technology, 20(1):53-67, 2001. 

	4. Richard Zanibbi, Kevin Novins, James Arvo and Katherine Zanibbi. 
	   ``Aiding Manipulation of Handwritten Mathematical Expressions through 
	   Style-Preserving Morphs'', In Proceedings of Graphics Interface 2001, 
	   Ottawa, Canada, June 2001, pp. 127-134. 

******************************************************************************/


// Steve's character recogniser!

#include <cstring>
#include <cstdlib>
#include <cstdio>
#include <cassert>
#include <pthread.h>

#include <iostream>
#include <fstream>

// Tcl/Tk includes
#include <tcl.h>
#include <tk.h>

// Smartboard includes
#include <util/Vec2.h>		// 2-D vectors
#include <util/Multiline.h>	// multilines
#include <rec/Recognizer.h>	// the stroke recogniser
#include <util/Box2.h>		// 2-D boxes.

// My includes
#include "charrec.h"
#include "MapWindow.h"
#include "segmenter.h"
#include "RecognitionResults.h"
#include "RecognitionPartitions.h"

// Global variables.
static CharRec CR;

	//This is a wrapper function to receive the TCL data
	//but most of recogniseStrokes will be handled elsewhere 
	//(currently charrec.cc)
int recogniseStrokesTCL(ClientData client_data,
		     Tcl_Interp* interp,
		     int argc,
		     const char *argv[])
{

		//Declare needed objects
	Segmenter segObj;
	RecognitionResults recObj;
	RecognitionPartitions partObj;
	Profile profObj;
	vector<int> highestKey;
	vector<vector<string> > names;
	vector<vector<double> > confs;
	vector<Multiline> strokeMultilines;
	static char buff[1024];

		//JZ: May 31, 2006
		//Recognise_this now takes 7 arguments (6: extra argument for the time  7:Flag for handdrawn/image).
	if (argc != 7)
	{
		sprintf(interp->result,"Wrong number of arguments in recognise_this in ffes.cc");
		return TCL_ERROR;
	}

	if(strcmp(argv[6], "handdrawn") == 0) {
		strokeMultilines = CR.convertTclDataToMultilines(argc, argv);
		CR.CombineOverlappingMultilines(strokeMultilines);
		segObj.recognizeStrokes(strokeMultilines, CR.getRecogniser(), profObj, recObj);
	} else {
			//Image stroke
		strokeMultilines = CR.convertTclImageDataToMultilines(argc, argv);
		segObj.recognizeStrokes(strokeMultilines, CR.getImageRecogniser(), profObj, recObj);
	}


	partObj.insertRecognitionPartitions(strokeMultilines.size(), recObj);
	partObj.getHighestPartitionData(highestKey, names, confs);
	string tclResult = CR.MakeRecognitionResultForTcl(highestKey, names);

		//JZ: Jun  8, 2006
		//This is a bit messy, but we cannot just copy tclResult straight into
		//the interpreter because the memory gets corrupted after tclResult gets
		//eliminted from the stack.
		//
		//Ideal would be to use dynamically allocated memory and a freeProc in
		//TclSetResult command, but want to get code working first.
	sprintf(buff, "%s", tclResult.c_str());
	interp->result = buff;
	return TCL_OK;
}


// R. Zanibbi, April 29, 2003: another new Tcl command to return a string
int CIT_symbols(ClientData client_data,
		      Tcl_Interp* interp,
		      int argc,
		      const char *argv[])
{
	char* pcNames = NULL;
	pcNames = CR.GetSymbolNames();
 	Tcl_SetResult(interp, pcNames, TCL_VOLATILE);
	delete pcNames; // "new" is used in GetSymbolNames()
 	return TCL_OK;
}


int my_init( Tcl_Interp *interp )
{

	if(Tcl_Init(interp) == TCL_ERROR)
		return TCL_ERROR;
	if(Tk_Init(interp) == TCL_ERROR)
		return TCL_ERROR;

  Tcl_CreateCommand( interp, "recognise_this", recogniseStrokesTCL, NULL, NULL);
  Tcl_CreateCommand( interp, "CIT_SymbolList", CIT_symbols, NULL, NULL);

  if (MapWindow_Init(interp) == TCL_ERROR) 
    return TCL_ERROR;
  
  // R. Zanibbi, March 15, 2002: use an environment variable so
  // that we can run FFES places other than the ffes_0.2 directory.
  //
  // R. Zanibbi April 23, 2004: changed path to interface file.
  char * FFESDir = getenv("FFES_DIR");
  char * InterfaceFile = "/src/Tk_Interface/Interface.tk";
  
  if (FFESDir == NULL) {
	  cerr << "ERROR: FFES_DIR needs to be defined in the environment.\n\tIt"
		  << " should point to the base directory of FFES (e.g. <path>/ffes_0.4)\n";
	  exit(1);
  }

  char MainTkInterfaceFile[256] = "";
  strcat(MainTkInterfaceFile,FFESDir);
  strcat(MainTkInterfaceFile,InterfaceFile);

  // R. Zanibbi, April 24, 2003: adding explicit test for FFES_DIR
  // value, telling user if they need to make a correction.
  ifstream InterfaceExists(MainTkInterfaceFile);
  if(! InterfaceExists)
  {
	  cerr << "ERROR: Could not find the main interface file (Interface.tk) for FFES at:\n"
		   << "    $(FFES_DIR)" << InterfaceFile << "\n\n"
		   << "    Perhaps the FFES_DIR environment variable is incorrectly defined.\n"
		   << "    Currently FFES_DIR is defined as: \n"
		   << "	   " << FFESDir << "\n\n"
		   << "    Please make sure that FFES_DIR points to the base directory of FFES.\n"
		   << "    (e.g. <path>/ffes_0.4)\n";
	  	
	  InterfaceExists.close();
	  exit(1);
  }
  
  InterfaceExists.close();

  Tcl_SetVar( interp, "tcl_rcFileName", MainTkInterfaceFile, TCL_GLOBAL_ONLY );
  return 0;
}


int main(int argc, char **argv)
{
  Tk_Main( argc, argv, my_init );
  return 0;
}
